/****************************************************************************

 Copyright (c) 1999,2000,2001 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 getdatasock.c  v 0.0.0 2000/01/11
                v 1.6.0 2000/03/24  - Set Receive buffer to TCPWINSIZE
                                      for the get option
                v 1.7.1 2000/04/04  - Variable cleaning (detected on IRIX)
                v 1.8.7 2000/05/28  - Change headers
                v 1.8.9 2000/08/08  - Add timestamp
                v 1.9.0 2000/08/21  - Use configure to help portage
                v 2.0.0 2000/03/19  - Use protocol version 2
                v 2.0.1 2001/04/19  - Correct indentation 
                                    - Port to IRIX

*****************************************************************************/
#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <errno.h>

#include <bbftp.h>
#include <client.h>
#include <client_proto.h>
#include <common.h>
#include <structures.h>

extern int     debug ;
extern int     timestamp ;
extern int     *myports ;
extern int     *mysockets ;
extern int      warning ;
extern int      sendwinsize ;
extern int      recvwinsize ;
extern struct   sockaddr_in myctladdr ;

int getdatasock(int nbsock, int *errcode) 
{
    int     i ;
    int     tcpwinsize ;
    int     on = 1 ;
#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX)
    int     addrlen ;
#else
    size_t  addrlen ;
#endif
    struct  linger li ;
    struct  sockaddr_in sck ;
    int     *mysockfree ;
    int     *myportfree;
    
    mysockfree = mysockets ;
    myportfree = myports ;
    sck = myctladdr ;
    for (i=0 ; i<nbsock ; i++) {
        sck.sin_port = 0;
        *mysockfree = socket ( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
        if ( *mysockfree < 0 ) {
            printmessage(stderr,CASE_ERROR,91,timestamp,"Cannot create data socket: %s\n",strerror(errno));
            *errcode = 91 ;
            return -1 ;
        }
        if ( setsockopt(*mysockfree,SOL_SOCKET, SO_REUSEADDR,(char *)&on,sizeof(on)) < 0 ) {
            printmessage(stderr,CASE_ERROR,92,timestamp,"Cannot set SO_REUSEADDR on data socket : %s\n",strerror(errno));
            *errcode = 92 ;
            return -1 ;
        }
        if ( setsockopt(*mysockfree,IPPROTO_TCP, TCP_NODELAY,(char *)&on,sizeof(on)) < 0 ) {
            printmessage(stderr,CASE_ERROR,93,timestamp,"Cannot set TCP_NODELAY on data socket : : %s\n",strerror(errno));
            *errcode = 93 ;
            return -1 ;
        }
        /*
        ** Send and receive are inversed because they are supposed
        ** to be the daemon char
        */
        tcpwinsize = 1024 * recvwinsize ;
        if ( setsockopt(*mysockfree,SOL_SOCKET, SO_SNDBUF,(char *)&tcpwinsize,sizeof(tcpwinsize)) < 0 ) {
            addrlen = sizeof(tcpwinsize) ;
            if ( getsockopt(*mysockfree,SOL_SOCKET,SO_SNDBUF,(char *)&tcpwinsize,&addrlen) < 0 ) {
                if ( warning) printmessage(stderr,CASE_WARNING,23,timestamp,"Unable to get send buffer size : %s\n",tcpwinsize,strerror(errno));
            } else {
                if ( warning) printmessage(stderr,CASE_WARNING,24,timestamp,"Send buffer cannot be set to %d Bytes, Value is %d Bytes\n",1024*recvwinsize,tcpwinsize);
            }
        }
        tcpwinsize = 1024 * sendwinsize ;
        if ( setsockopt(*mysockfree,SOL_SOCKET, SO_RCVBUF,(char *)&tcpwinsize,sizeof(tcpwinsize)) < 0 ) {
            addrlen = sizeof(tcpwinsize) ;
            if ( getsockopt(*mysockfree,SOL_SOCKET,SO_RCVBUF,(char *)&tcpwinsize,&addrlen) < 0 ) {
                if ( warning) printmessage(stderr,CASE_WARNING,25,timestamp,"Unable to get receive buffer size : %s\n",tcpwinsize,strerror(errno));
            } else {
                if ( warning) printmessage(stderr,CASE_WARNING,26,timestamp,"Receive buffer cannot be set to %d Bytes, Value is %d Bytes\n",1024*sendwinsize,tcpwinsize);
            }
        }    
        li.l_onoff = 1 ;
        li.l_linger = 1 ;
        if ( setsockopt(*mysockfree,SOL_SOCKET,SO_LINGER,(char *)&li,sizeof(li)) < 0 ) {
            printmessage(stderr,CASE_ERROR,94,timestamp,"Cannot set SO_LINGER on data socket : %s\n",strerror(errno));
            *errcode = 94 ;
            return -1 ;
        }
        if ( bind(*mysockfree, (struct sockaddr *) &sck ,sizeof(sck))  < 0) {
            printmessage(stderr,CASE_ERROR,95,timestamp,"Cannot bind on data socket : %s\n",strerror(errno));
            *errcode = 95 ;
            return -1 ;
        }
        addrlen = sizeof(sck) ;
        if (getsockname(*mysockfree,(struct sockaddr *)&sck, &addrlen) < 0) {
            printmessage(stderr,CASE_ERROR,96,timestamp,"Cannot getsockname on data socket : %s\n",strerror(errno));
            *errcode = 96 ;
            return -1 ;
           }
        if (listen(*mysockfree, 1) < 0) {
            printmessage(stderr,CASE_ERROR,97,timestamp,"Cannot listen on data socket : %s\n",strerror(errno));
            *errcode = 97 ;
            return -1 ;
        }
        *myportfree++ =  sck.sin_port ;
        if ( debug ) printmessage(stdout,CASE_NORMAL,0,timestamp,"Port number = %d\n", sck.sin_port) ;
        mysockfree++ ;
    }
    return 0 ;
}
