/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 bbftpd.c v 0.0.0  1999/11/24
          v 1.3.0  2000/03/16   - Version in openlog
          v 1.4.0  2000/03/22   - Modify the connection protocole
                                  in order to send the crypt type
                                  after the connection.
          v 1.5.0  2000/03/24   - Change the wait timer in default
          v 1.6.0  2000/03/24   - Change the main loop to make it more
                                    modular
          v 1.6.1  2000/03/24   - Portage on OSF1
                                - Make the control socket non blocking
          v 1.8.0  2000/04/14   - Introduce RSA Cryptage
          v 1.8.6  2000/05/21   - Allow to run with inetd
          v 1.8.7  2000/05/24   - Introduce version.h and config.h
          v 1.8.10 2000/08/11   - Portage to Linux
          v 1.9.0  2000/08/18   - Use configure to help portage
                                - default time out set to 900 s
          v 1.9.3  2000/10/12   - Add -b and -w option in order to overwrite
                                  the fixed values
          v 1.9.4  2000/10/16   - Make all sockets blocking in order
                                  to prevent the error in case of lack 
                                  of memory
                                - Supress %m
          v 2.0.0  2001/03/26   - Protocol V2 implementation
          v 2.0.1  2001/04/23   - Correct indentation
                                - Port to IRIX
          v 2.0.2  2001/05/07   - Add debug option for RFIO
          v 2.1.0  2001/05/28   - Add private authentication
                                - Correct syslog level
                                - Add -l option

*****************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#include <utime.h>

#include <bbftpd.h>
#include <common.h>
#include <config.h>
#include <daemon.h>
#include <daemon_proto.h>
#include <status.h>
#include <structures.h>
#include <version.h>

#include <openssl/rsa.h>

#define OPTIONS    "bd:fl:m:svw:"
/*
** Common variables for BBFTP protocole version 1 and 2
*/
/*
** be_daemon :
**      That is the way the daemon is going to be run.
**      0 means run by inetd (that can be done trought 
**      a wrapper)                                      (run as bbftpd)
**      1 means that the control socket are stdin and 
**      stdout                                          (run as bbftpd -s)
**      2 means running in background                   (run as bbftpd -b)
**      Default 0
*/
int        be_daemon = 0 ;
/*
** daemonchar:
**      String used to initialize openlog
*/
char        daemonchar[50] ;
/*
** fixeddataport :
**      If set to 1 then the server will use CONTROLPORT-1 
**      to use as local port number for the data socket. 
**      It is useful to set it to 1 if the server is 
**      running behing a firewall.                      (run as bbftpd)
**      If set to 0 then the server will not fixe the 
**      local data port                                 (run as bbftpd -f)
**      Default 1
*/
int     fixeddataport = 1 ;
/*
** state :
**      Define the state of the server. values differ depending
**      of the BBFTP protocol version
*/
int        state ;                         
/*
** newcontrolport :
**      Define the control port to listen on
*/
int        newcontrolport ; 
/*
** currentusername :
**      Define the local username
*/
char currentusername[MAXLEN] ;
/*
** myrsa :
**      Define the local location where is store the key pair
*/
RSA        *myrsa ;
/*
** his_addr :
**      Remote address (the client)
*/
struct sockaddr_in his_addr;
/*
** ctrl_addr :
**      Local adresse (the server)
*/
struct sockaddr_in ctrl_addr;
/*
** protocolversion :
**      Set to 0 before any PROT message exchange, that also means
**      that no children have been started
*/
int     protocolversion ;
int     protocolmin = 1  ;
int     protocolmax = 2  ;
/*
** fatherpid :
**      The pid of the process keepng the control connection
*/
pid_t    fatherpid ;
/* 
** flagsighup:
**      Set to one when a SIGHUP is received
**/
int    flagsighup = 0 ;
/*
** killdone:
**      Set to one when the father has send a kill to all chidren
*/
int    killdone= 0 ;
/*
** childendinerror:
**      Set to one when one child has been detected in error in order
**      to prevent father to send multiple answer
*/
int    childendinerror = 0 ;
/*
** unlinkfile:
**      Set to one when father has to unlink the file to store. Used
**      in case of child ended incorrectly (kill -9 for example)
*/
int    unlinkfile = 0  ;
/*
** debug:
**      Set to one for RFIO debug. Valid only with the -b option
**      set to the value needed for RFIO_TRACE
*/
int     debug = 0 ;
/*
** castfd:
**      CASTOR file descriptor use with RFIO and CASTOR
*/
int     castfd ;
/*
** castfilename:
**      CASTOR/SHIFT real file name
*/
char    *castfilename = NULL ;
/*
** End Common variables for BBFTP protocole version 1 and 2 ********************
*/
/*
** Variables for BBFTP protocole version 1 
*/
/*
** msgsock :
**      Define control socket for the server
*/
int     msgsock ;
/*
** currentfilename :
**      Define the file we are working on 
*/
char    currentfilename[MAXLENFILE];
/*
** pid_child :
**      Array to store the children pid
*/
pid_t pid_child[MAXPORT] ;
/*
** End Variables for BBFTP protocole version 1 *********************************
*/
/*
** Variables for BBFTP protocole version 2
*/
/*
** incontrolsock :
**      Define the control socket for reading
** outcontrolsock :
**      Define the control socket for writing
**
**      For be_daemon equal to 0 or 2 incontrolsock = outcontrolsock
*/
int     incontrolsock ;
int     outcontrolsock ;
/*
** curfilename :
**      Define the pointer to the current file
*/
char    *curfilename = NULL ;
/*
** curfilenamelen :
**      Define the length of the memory used by curfilename
*/
int     curfilenamelen ;
/*
** realfilename :
**      Define the pointer to the real file (= curfilename if TROPT_TMP not
**      set)
*/
char    *realfilename = NULL ;
/*
** mychildren :
**      Pointer to the first pid of children
*/
int     *mychildren = NULL ;
/*
** nbpidchid :
**      Number of pid pointed by mychildren 
*/
int     nbpidchild ;
/*
** myports :
**      Pointer to the first port
*/
int     *myports = NULL ;
/*
** readbuffer :
**      Pointer to the readbuffer
*/
int     *readbuffer = NULL ;
/*
** compbuffer :
**      Pointer to the compression buffer
*/
int     *compbuffer = NULL ;
/*
** myumask :
**      Umask for the bbftpd process, the default will be 022 which
**      means that group and other write permissions are denied
**      This can be changed by the bbftp umask command
**/
int     myumask = 022 ;
/*
** Parameters describing the transfer
*/
int     transferoption ;
int     filemode ;
char    lastaccess[9] ;
char    lastmodif[9] ;
int     sendwinsize ;        
int     recvwinsize ;        
int     buffersizeperstream ;
int     requestedstreamnumber ;
my64_t  filesize ;
int     mycos = -1 ;
/*
** maxstreams :
**      That is to fix the maximum number of stream the deamon will accept.
**      (run as bbftpd -m 25)
**      Default 25
*/
int     maxstreams = 25 ;
/*
** End Variables for BBFTP protocole version 2 *********************************
*/


main (argc,argv,envp)
    int     argc ;
    char    **argv ;
    char    **envp ;
{
    extern char *optarg;
    extern int optind, opterr, optopt;

#if defined(SUNOS) || defined(_HPUX_SOURCE) || defined(IRIX)
    int        addrlen ;
#else
    size_t        addrlen ;
#endif
    struct  timeval    wait_timer;
    fd_set  selectmask ; 
    int        nfds ; 
    int        retcode ;
    int        i, j ;
    struct  message *msg ;
    char    buffer[MINMESSLEN] ;
    char    logmessage[1024] ;
    char    rfio_trace[20] ;
    
    sprintf(daemonchar,"bbftpd v%s",VERSION) ;
    openlog(daemonchar, LOG_PID | LOG_NDELAY, BBFTPD_FACILITY);
    /*
    ** Set the log mask to BBFTPD_EMERG (0) 
    */
    setlogmask(LOG_UPTO(BBFTPD_EMERG));
    /*
    ** Initialize variables
    */
    protocolversion = 0 ;
    
    newcontrolport = CONTROLPORT ;
    opterr = 0 ;
    while ((j = getopt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'v' :{
                printf("bbftpd version %s\n",VERSION) ;
                printf("Compiled with  :   default port %d\n",CONTROLPORT) ;
                printf("                   default maximum streams = %d \n",maxstreams) ;
#ifdef PRIVATE_AUTH
                printf("                   private authentication \n") ;
#endif
                exit(0) ;
            }
        }
    }
    /*
    ** Look at the loglevel option 
    */
    i = 0 ;
    opterr = 0 ;
    optind = 1 ;
    while ((j = getopt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'l' :{
                for ( i=0 ; i< strlen(optarg) ; i++ ) {
                    optarg[i] = toupper(optarg[i]) ;
                }
                i = 0 ;
                if ( !strcmp(optarg,"EMERGENCY") ) {
                    i = BBFTPD_EMERG;
                } else if ( !strcmp(optarg,"ALERT") ) {
                    i = BBFTPD_ALERT;
                } else if ( !strcmp(optarg,"CRITICAL") ) {
                    i = BBFTPD_CRIT;
                } else if ( !strcmp(optarg,"ERROR") ) {
                    i = BBFTPD_ERR;
                } else if ( !strcmp(optarg,"WARNING") ) {
                    i = BBFTPD_WARNING;
                } else if ( !strcmp(optarg,"NOTICE") ) {
                    i = BBFTPD_NOTICE;
                } else if ( !strcmp(optarg,"INFORMATION") ) {
                    i = BBFTPD_INFO;
                } else if ( !strcmp(optarg,"DEBUG") ) {
                    i = BBFTPD_DEBUG;
                }
                if ( i > 0 ) {
                    setlogmask(LOG_UPTO(i));
                }
                break ;
            }
        }
    }
    syslog(BBFTPD_NOTICE,"Starting bbftpd") ;
    opterr = 0 ;
    optind = 1 ;
    while ((j = getopt(argc, argv, OPTIONS)) != -1) {
        switch (j) {
            case 'b' :{
                if ( be_daemon != 0 ) {
                    syslog(BBFTPD_ERR,"-b and -s options are incompatibles") ;
                    exit(1) ;
                }
                be_daemon = 2 ;
                break ;
            }
            case 'd' :{
                sscanf(optarg,"%d",&i) ;
                if ( i > 0 ) debug = i ;
                break ;
            }
            case 'f' :{
                fixeddataport = 0 ;
                break ;
            }
            case 'm' :{
                sscanf(optarg,"%d",&i) ;
                if ( i > 0 ) maxstreams = i ;
                break ;
            }
            case 's' :{
                if ( be_daemon != 0 ) {
                    syslog(BBFTPD_ERR,"-b and -s options are incompatibles") ;
                    exit(1) ;
                }
#ifdef PRIVATE_AUTH
                syslog(BBFTPD_ERR,"-s option cannot be used with private authentication") ;
                exit(1) ;                
#endif
                be_daemon = 1 ;
                break ;
            }
            case 'w' :{
                sscanf(optarg,"%d",&newcontrolport) ;
                break ;
            }
            default : {
                break ;
            }
        }
    }
    /*
    ** Get 5K for castfilename in order to work with this @#! CASTOR
    ** software (even if we are not using it)
    */
    if ( (castfilename = (char *) malloc (5000)) == NULL ) {
        /*
        ** Starting badly if we are unable to malloc 5K
        */
        syslog(BBFTPD_ERR,"No memory for this fucking CASTOR : %s",strerror(errno)) ;
        fprintf(stderr,"No memory for this fucking CASTOR : %s\n",strerror(errno)) ;
        exit(1) ;
    }
    /*
    ** Reset debug to zero if -b option is not present
    */
    if ( (be_daemon != 2) ) debug = 0 ;
    /*
    ** Check if be_daemon = 0 and in this case reset the 
    ** control port to CONTROLPORT
    */
    if ( be_daemon == 0 ) newcontrolport = CONTROLPORT ;
                
    if ( be_daemon == 2 ) {
        /*
        ** Run as a daemon 
        */
        do_daemon(argc, argv, envp);
        /*
        ** Check for debug
        */
        if ( debug != 0 ) {
            sprintf(rfio_trace,"RFIO_TRACE=%d",debug) ;
            retcode = putenv(rfio_trace) ;
            if ( retcode == 0 ) {
                /*
                ** reopen stdout to a file like /tmp/bbftpd.rfio.trace.pid
                */
                close(STDOUT_FILENO) ;
                sprintf(logmessage,"/tmp/bbftp.rfio.trace.level.%d.%d",debug,getpid()) ;
                (void) freopen(logmessage,"w",stdout) ;
            }
        }
    } else if ( be_daemon == 1 ) {
        /*
        ** Run by a remote user
        */
        /*
        ** Get the username 
        */
        struct passwd *result ;
        if ( (result = getpwuid(getuid())) == NULL ) {
            syslog(BBFTPD_WARNING,"Error getting username") ;
            sprintf(currentusername,"UID %d",getuid()) ;
        } else {
            strcpy(currentusername,result->pw_name) ;
        }
        syslog(BBFTPD_NOTICE,"bbftpd started by : %s",currentusername) ;
        /*
        ** Set the control sock to stdin and stdout
        */
        incontrolsock = 0  ;
        outcontrolsock = 1 ;
        /*
        ** As we cannot rely on the env variables to know the
        ** remote host, we are going to wait on an undefined 
        ** port, send the MSG_LOGGED_STDIN and the port number
        ** and wait for a connection...
        */
        checkfromwhere() ;
        syslog(BBFTPD_NOTICE,"bbftpd started by : %s from %s",currentusername,inet_ntoa(his_addr.sin_addr)) ;
    } else {
        char    buffrand[NBITSINKEY] ;
        struct timeval tp ;
        unsigned int seed ;
        /*
        ** Load the error message from the crypto lib
        */
        ERR_load_crypto_strings() ;
        /*
        ** Initialize the buffrand buffer which is giong to be used to initialize the 
        ** random generator
        */
        /*
        ** Take the usec to initialize the random session
        */
        gettimeofday(&tp,NULL) ;
        seed = tp.tv_usec ;
        srandom(seed) ;
        for (i=0; i < sizeof(buffrand) ; i++) {
            buffrand[i] = random() ;
        }
        /*
        ** Initialize the random generator
        */
        RAND_seed(buffrand,NBITSINKEY) ;
        incontrolsock = 0  ;
        outcontrolsock = incontrolsock ;
        msgsock = incontrolsock ;
        /* 
        ** set the state
        */
        state = S_PROTWAIT ;
    }
    fatherpid = getpid() ;
    if ( be_daemon == 2 || be_daemon == 0 ) {
        /* Get the remote address */
        addrlen = sizeof(his_addr);
        if (getpeername(incontrolsock, (struct sockaddr *) &his_addr, &addrlen) < 0) {
            syslog(BBFTPD_ERR, "getpeername (%s): %s", argv[0],strerror(errno));
            exit(1);
        }
        addrlen = sizeof(ctrl_addr);
        if (getsockname(incontrolsock, (struct sockaddr *) &ctrl_addr, &addrlen) < 0) {
            syslog(BBFTPD_ERR, "getsockname (%s): %s", argv[0],strerror(errno));
            exit(1);
        }
        syslog(BBFTPD_NOTICE,"Getting new bbftp connexion from : %s",inet_ntoa(his_addr.sin_addr)) ;
        /*
        ** Send the encryption supported 
        */
        sendcrypt() ;
        /* 
        ** set the state
        */
        state = S_CONN ;
        nfds = sysconf(_SC_OPEN_MAX) ;
        FD_ZERO(&selectmask) ;
        FD_SET(incontrolsock,&selectmask) ;
        wait_timer.tv_sec  = 10 ;
        wait_timer.tv_usec = 0 ;
        if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
            syslog(BBFTPD_ERR,"Select error : %s",strerror(errno)) ;
            exit(1) ;
        } else if ( retcode == 0 ) {
            syslog(BBFTPD_ERR,"Time OUT ") ;
            exit(1) ;
        } else {
            if ( (retcode = readmessage(incontrolsock,buffer,MINMESSLEN,CONTROLSOCKTO)) < 0 ) {
                syslog(BBFTPD_ERR,"Error reading MSG_LOG ") ;
                exit(1) ;
            }
            msg = (struct message *) buffer ;
#ifndef PRIVATE_AUTH
            if ( msg->code == MSG_LOG ) {
                /*
                ** It seems that it is the message we were waiting
                ** so lets decrypt 
                */
                if ( loginsequence() < 0 ) {
                    /*
                    ** The login failed, do not wait for a new one
                    */
                    exit(1) ;
                }
                state = S_PROTWAIT ;
                sprintf(logmessage,"bbftpd version %s : OK",VERSION) ;
                reply(MSG_OK,logmessage) ;
            } else {
                syslog(BBFTPD_ERR,"Unkown message in connected state : %d",msg->code) ;
                reply(MSG_BAD,"Unkown message in connected state") ;
                exit(1) ;
            }
#else
            if ( msg->code == MSG_PRIV_LOG ) {
                if ( bbftpd_private_receive_connection(msg->msglen) < 0 ) {
                    /*
                    ** The login failed, do not wait for a new one
                    */
                    exit(1) ;
                }
                state = S_PROTWAIT ;
                sprintf(logmessage,"bbftpd version %s : OK",VERSION) ;
                reply(MSG_OK,logmessage) ;
            } else {
                syslog(BBFTPD_ERR,"Unkown message in connected state : %d",msg->code) ;
                reply(MSG_BAD,"Unkown message in connected state") ;
                exit(1) ;
            }
#endif
        }
    }

    /*
    ** At this stage we are in the S_PROTWAIT state
    */
    nfds = sysconf(_SC_OPEN_MAX) ;
    FD_ZERO(&selectmask) ;
    FD_SET(incontrolsock,&selectmask) ;
    wait_timer.tv_sec  = 10 ;
    wait_timer.tv_usec = 0 ;
    if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
        syslog(BBFTPD_ERR,"Select error in S_PROTWAIT state : %s",strerror(errno)) ;
        exit(1) ;
    } else if ( retcode == 0 ) {
        syslog(BBFTPD_ERR,"Time OUT in S_PROTWAIT state") ;
        exit(1) ;
    } else {
        if ( (retcode = readmessage(incontrolsock,buffer,MINMESSLEN,CONTROLSOCKTO)) < 0 ) {
            syslog(BBFTPD_ERR,"Error reading in S_PROTWAIT state") ;
            exit(1) ;
        }
        msg = (struct message *) buffer ;
        if ( msg->code == MSG_PROT ) {
            /*
            ** The client is using bbftp v2 protocol or higher
            */
            if ( checkprotocol() < 0 ) {
                exit_clean() ;
                exit(1) ;
            }
            syslog(BBFTPD_INFO,"Using bbftp protocol version %d",protocolversion) ;
            state = S_LOGGED ;
            /*
            ** Initialize the variables
            */
            mychildren = NULL ;
            nbpidchild = 0 ;
            curfilename = NULL ;
            curfilenamelen = 0 ;
        } else {
            /*
            ** This is a bbftp v1 client 
            */
            protocolversion = 1 ;
            syslog(BBFTPD_INFO,"Using bbftp protocol version 1") ;
            state = S_LOGGED ;
            /*
            ** So set up the v1 handlers
            */
            if ( set_signals_v1() < 0 ) {
                exit(1) ;
            }
            /*
            ** Initialize the pid array
            */
            for ( i=0 ; i< MAXPORT ; i++) {
                pid_child[i] = 0 ;
            }
            /*
            ** As we have already read the message 
            */
            if ( readcontrol(msg->code,msg->msglen) < 0 ) { 
                clean_child() ;
                exit_clean() ;
                exit(0) ;
            }
        }
    }
    if ( protocolversion == 1 ) goto loopv1 ;
    if ( protocolversion == 2 ) goto loopv2 ;
    syslog(BBFTPD_ERR,"Unknown protocol version %d",protocolversion) ;
    exit(1) ;
/*
** Loop for the v2 protocol 
*/
loopv2:
    /*
    ** Set up v2 handlers
    */
    if ( bbftpd_setsignals() < 0 ) {
        exit(1) ;
    }
    /*
    ** Set the umask ; first unset it and then reset to the default value
    */
    umask(0) ;
    umask(myumask) ;
    for (;;) {
        /*
        ** Initialize the selectmask
        */
        nfds = sysconf(_SC_OPEN_MAX) ;
        FD_ZERO(&selectmask) ;
        FD_SET(incontrolsock,&selectmask) ;
        /*
        ** Depending on the state set a timer or not 
        */
        switch (state) {
            case S_WAITING_STORE_START :
            case S_WAITING_FILENAME_STORE :
            case S_WAITING_FILENAME_RETR : {
                /*
                ** Timer of 10s between XX_V2 and FILENAME_XX
                */
                wait_timer.tv_sec  = 10 ;
                wait_timer.tv_usec = 0 ;
                break ;
            }
            
            case S_SENDING : 
            case S_RECEIVING : {
                /*
                ** No timer while receiving or sending
                */
                wait_timer.tv_sec  = 0 ;
                wait_timer.tv_usec = 0 ;
                break ;
            }
            default : {
                /*
                ** Timer of 900s between commands
                */
                wait_timer.tv_sec  = 900 ;
                wait_timer.tv_usec = 0 ;
                break ;
            }
        }
        if ( (retcode = select(nfds,&selectmask,0,0,(wait_timer.tv_sec == 0) ? NULL : &wait_timer) ) == -1 ) {
            if ( errno != EINTR ) {
                syslog(BBFTPD_ERR,"Select error : %s",strerror(errno)) ;
            }
        } else if ( retcode == 0 ) {
            syslog(BBFTPD_ERR,"Time OUT ") ;
            if ( state == S_WAITING_STORE_START ) {
                bbftpd_storeunlink(realfilename) ;
            }
            clean_child() ;
            exit_clean() ;
            exit(0) ;
        } else {
            /*
            ** At this stage we can only receive a command
            */
            if ( (retcode = readmessage(incontrolsock,buffer,MINMESSLEN,CONTROLSOCKTO)) < 0 ) {
                if ( state == S_WAITING_STORE_START  || state == S_RECEIVING) {
                    bbftpd_storeunlink(realfilename) ;
                    sleep(5) ;
                }
                clean_child() ;
                exit_clean() ;
                exit(0) ;
            }
            if ( bbftpd_readcontrol(msg->code,msg->msglen) < 0 ) { 
                clean_child() ;
                exit_clean() ;
                exit(0) ;
            } else {
            }
        }
    }
/*
** Loop for the v1 protocol 
*/
loopv1:
    for (;;) {
        /*
        ** Initialize the selectmask
        */
        nfds = sysconf(_SC_OPEN_MAX) ;
        FD_ZERO(&selectmask) ;
        FD_SET(incontrolsock,&selectmask) ;
        /*
        ** Depending on the state set a timer or not 
        */
        switch (state) {
            case S_SENDING : 
            case S_RECEIVING : {
                /*
                ** No timer while receiving or sending
                */
                wait_timer.tv_sec  = 0 ;
                wait_timer.tv_usec = 0 ;
                break ;
            }
            default : {
                /*
                ** Timer of 900s between commands
                */
                wait_timer.tv_sec  = 900 ;
                wait_timer.tv_usec = 0 ;
                break ;
            }
        }
        if ( (retcode = select(nfds,&selectmask,0,0,(wait_timer.tv_sec == 0) ? NULL : &wait_timer) ) == -1 ) {
            if ( errno != EINTR ) {
                syslog(BBFTPD_ERR,"Select error : %s",strerror(errno)) ;
            }
        } else if ( retcode == 0 ) {
            syslog(BBFTPD_ERR,"Time OUT ") ;
            clean_child() ;
            exit_clean() ;
            exit(0) ;
        } else {
            /*
            ** At this stage we can only receive a command
            */
            if ( (retcode = readmessage(incontrolsock,buffer,MINMESSLEN,CONTROLSOCKTO)) < 0 ) {
                clean_child() ;
                exit_clean() ;
                exit(0) ;
            }
            if ( readcontrol(msg->code,msg->msglen) < 0 ) { 
                clean_child() ;
                exit_clean() ;
                exit(0) ;
            } else {
            }
        }
    }
}

void clean_child() 
{
    int    *pidfree ;
    int    i ;
    
    if ( protocolversion == 0 ) return ;
    if ( protocolversion == 1 ) {
        if ( killdone == 0 ) {
            killdone = 1 ;
            for ( i=0 ; i<MAXPORT ; i++) {
                if ( pid_child[i] != 0 ) {
                    syslog(BBFTPD_DEBUG,"Killing child %d",pid_child[i]) ;
                    kill(pid_child[i],SIGKILL) ;
                }
            }
        }
        return ;
    }
    if ( protocolversion == 2 ) {
        if ( killdone == 0 ) {
            killdone = 1 ;
            pidfree = mychildren ;
            for ( i=0 ; i<nbpidchild ; i++) {
                if ( *pidfree != 0 ) {
                    syslog(BBFTPD_DEBUG,"Killing child %d",*pidfree) ;
                    kill(*pidfree,SIGKILL) ;
                }
                pidfree++ ;
            }
        }
        return ;
    }
}

void exit_clean() 
{
    
    switch (state) {
        case S_CONN : {
            return ;
        }
        case S_LOGGED :
        case S_PROTWAIT :
        case S_WAITING_FILENAME_STORE :
        case S_WAITING_STORE_START :
        case S_SENDING:
        case S_RECEIVING : {
            syslog(BBFTPD_NOTICE,"User %s disconnected",currentusername) ;
            return ;
        }
        default :{
            return ;
        }
    }
}

my64_t convertlong(my64_t v) {
    struct bb {
        int    fb ;
        int sb ;
    } ;
    struct bb *bbpt ;
    int     tmp ;
    my64_t    tmp64 ;
    
    tmp64 = v ;
    bbpt = (struct bb *) &tmp64 ;
    tmp = bbpt->fb ;
    bbpt->fb = ntohl(bbpt->sb) ;
    bbpt->sb = ntohl(tmp) ;    
    return tmp64 ;
}
