/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN:
        0  Keep the connection open (does not mean that the directory has been
           successfully created)
        -1 Tell the calling program to close the connection
  
 bbftpd_mkdir.c v 2.0.0  2001/02/12 - Creation of the routine.
                v 2.0.1  2001/04/23 - Correct indentation
                                    - Port to IRIX
                v 2.1.0  2001/06/01 - Reorganise routines as in bbftp_
                                      
 *****************************************************************************/

#include <netinet/in.h>
#include <stdio.h>
#include <syslog.h>
#include <utime.h>

#include <bbftpd.h>
#include <common.h>
#include <daemon.h>
#include <daemon_proto.h>
#include <structures.h>

extern int transferoption ;

int bbftpd_mkdir(int sock,int msglen) 
{
    char    *buffer ;
    char    *logmessage ;
    struct  mess_dir *msg_dir ;
    char    *dirname ;
    int     recursif ;
    int     retcode ;

    if ( (buffer = (char *) malloc (msglen+1) ) == NULL ) {
        syslog(BBFTPD_ERR,"Unable to malloc space for directory name (%d)",msglen) ;
        reply(MSG_BAD,"Unable to malloc space for directory name") ;
        return 0 ;
    }
    if ( (logmessage = (char *) malloc (msglen+1024) ) == NULL ) {
        syslog(BBFTPD_ERR,"Unable to malloc space for logmessage ") ;
        reply(MSG_BAD,"Unable to malloc space for logmessage") ;
        free(buffer) ;
        return 0 ;
    }
    /*
    ** Read the characteristics of the directory
    */
    if ( readmessage(sock,buffer,msglen,CONTROLSOCKTO) < 0 ) {
        /*
        ** Error ...
        */
        syslog(BBFTPD_ERR,"Error reading directory name") ;
        free(buffer) ;
        free(logmessage) ;
        return -1 ;
    }
    buffer[msglen] = '\0' ;
    msg_dir = (struct mess_dir *) buffer ;
    transferoption  = msg_dir->transferoption ;
    dirname = msg_dir->dirname ;
    if ( (transferoption & TROPT_DIR ) == TROPT_DIR ) {
        recursif = 1 ;
    } else {
        recursif = 0 ;
    }
    if ( (retcode = bbftpd_storemkdir(dirname,logmessage,recursif)) < 0 ) {
        reply(MSG_BAD_NO_RETRY,logmessage) ;
        free(buffer) ;
        free(logmessage) ;
        return 0 ;
    } else if (retcode > 0 ) {
        reply(MSG_BAD,logmessage) ;
        free(buffer) ;
        free(logmessage) ;
        return 0 ;
    } else {
        sprintf(logmessage,"Directory %s created",dirname) ;
        reply(MSG_OK,logmessage) ;
        free(buffer) ;
        free(logmessage) ;
        return 0 ;
    }
}
